/*
 * io.c - poll() based I/O loop
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <poll.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
//#include "log.h"
#include "failures.h"
#include "io.h"
#include "prioq.h"
#include "timeout.h"

#define IO_PARANOIA	1
#define IO_START_SIZE	32

#define ROUND_UP(a, b)	(((a)+((b)-1)/(b))*(b))

/*
 * fd_desc are allocated dynamically when a new fd is added to the set.
 */
struct fd_desc {
	int 		fd;		/* fd number */
	unsigned int	idx_to_fdset;	/* index of fdset */
	short		events;		/* event mask */
	void		*opaque;	/* opaque data passed to callback */
	iocallback_t	callback;	/* handler which knows about i/o enough */
	int		deleted_flag;	/* deletion marker */
};

#ifdef IO_PARANOIA
static void check_fd(struct io_context *io, int fd)
{
	assert(fd >= 0);
	assert(io->desc[fd] != NULL);
	assert(io->desc[fd]->fd == fd);
	assert(io->desc[fd]->idx_to_fdset >= 0);
	assert(io->desc_size > io->desc[fd]->idx_to_fdset);
	assert(io->fdset[io->desc[fd]->idx_to_fdset].fd == fd);
	/* check that only EVENT_READ or EVENT_WRITE are set */
	assert(!(io->desc[fd]->events & ~(EVENT_READ|EVENT_WRITE)));
	if (io->desc[fd]->events & EVENT_READ)
		assert(io->fdset[io->desc[fd]->idx_to_fdset].events &
				(POLLIN|POLLPRI));
	if (io->desc[fd]->events & EVENT_WRITE)
		assert(io->fdset[io->desc[fd]->idx_to_fdset].events &
				POLLOUT);
	assert(!(io->fdset[io->desc[fd]->idx_to_fdset].events &
			~(POLLIN|POLLPRI|POLLOUT)));
}
#endif

struct io_context *io_new()
{
	size_t i;
	struct io_context *io;

	io = xmalloc0(sizeof(*io));
	io->fdset = xmalloc0(sizeof(struct pollfd) * IO_START_SIZE);
	io->nfds_size = IO_START_SIZE;
	io->nfds_used = 0;
	io->desc = xmalloc0(sizeof(struct fd_desc *) * IO_START_SIZE);
	io->desc_size = IO_START_SIZE;
	io->loop_running = 1;
	io->fd_deleted_flag = 0;
	io->reduse_mem_flag = 0;
	io->current_desc = -1;
	for (i = 0; i < IO_START_SIZE; i++) {
		io->fdset[i].fd = -1;
		io->fdset[i].events = 0;
		io->fdset[i].revents = 0;
		io->desc[i] = NULL;
	}
	return io;
}

void io_free(struct io_context **io)
{
	int i;

	if (!io || *io == NULL)
		xerror(1, "io_free: io NULL");

#ifdef IO_PARANOIA
	for (i = 0; i < (*io)->nfds_used; i++)
		check_fd(*io, (*io)->fdset[i].fd);
#endif
	for (i = 0; i < (*io)->desc_size; i++) {
		if ((*io)->desc[i] != NULL) {
			check_fd(*io, i);
			xfree((*io)->desc[i]);
		}
	}
	xfree((*io)->desc);
	xfree((*io)->fdset);
	xfree((*io));
	*io = NULL;
}

int io_add(struct io_context *io, int fd, short events, iocallback_t callback, void *opaque)
{
	size_t i;
	struct fd_desc *d;
	size_t new_size;

	if (!io || !callback || fd < 0)
		xerror(1, "io_add: invalid arguments fd=%d callback=%p", fd, callback);

	if (fd >= io->desc_size) {
		new_size = io->desc_size;
		while (new_size <= fd)
			new_size *= 2;
		io->desc = xrealloc0(io->desc,
				io->desc_size * sizeof(struct fd_desc*),
				new_size * sizeof(struct fd_desc*));
		for (i = io->desc_size; i < new_size; i++)
			io->desc[i] = NULL;
		io->desc_size = new_size;
	}
	if (io->nfds_used == io->nfds_size) {
		io->fdset = xrealloc0(io->fdset,
				io->nfds_size * sizeof(struct pollfd),
				2 * io->nfds_size * sizeof(struct pollfd));
		for (i = io->nfds_size; i < io->nfds_size * 2; i++) {
			io->fdset[i].fd = -1;
			io->fdset[i].events = 0;
			io->fdset[i].revents = 0;
		}
		io->nfds_size *= 2;
	}
	if (io->desc[fd] == NULL || io->desc[fd]->deleted_flag) {
		if (io->desc[fd] == NULL)
			d = xmalloc0(sizeof(*d));
		else
			d = io->desc[fd];
		d->fd = fd;
		d->events = events & (EVENT_READ|EVENT_WRITE);
		//d->events |= (events & (EVENT_READ|EVENT_WRITE));
		d->opaque = opaque;
		d->callback = callback;
		d->deleted_flag = 0;
		d->idx_to_fdset = io->nfds_used++;
		io->fdset[d->idx_to_fdset].fd = fd;
		io->fdset[d->idx_to_fdset].revents = 0;
		if (events & EVENT_READ)
			io->fdset[d->idx_to_fdset].events = POLLIN|POLLPRI;
		if (events & EVENT_WRITE)
			io->fdset[d->idx_to_fdset].events = POLLOUT;
		io->desc[fd] = d;
	} else
		xerror(1, "io_add: fd=%d already exists in set!", fd);
	return 0;
}

void io_change(struct io_context *io, int fd, short events)
{
	if (!io)
		xerror(1, "io_change: fd=%d io NULL", fd);

	if (fd < 0)
		xerror(1, "io_change: invalid argument fd=%d", fd);

	if (io->desc[fd] == NULL)
		xerror(1, "io_change: fd=%d is not in the set", fd);

#ifdef IO_PARANOIA
	check_fd(io, fd);
#endif
	if (events & EVENT_READ) {
		if (!(io->desc[fd]->events & EVENT_READ)) {
			io->fdset[io->desc[fd]->idx_to_fdset].events |=
							(POLLIN | POLLPRI);
			io->desc[fd]->events |= EVENT_READ;
		}
	} else {
		if (io->desc[fd]->events & EVENT_READ) {
			io->fdset[io->desc[fd]->idx_to_fdset].events &=
							~(POLLIN|POLLPRI);
			io->desc[fd]->events &= ~EVENT_READ;
		}
	}

	if (events & EVENT_WRITE) {
		if (!(io->desc[fd]->events & EVENT_WRITE)) {
			io->fdset[io->desc[fd]->idx_to_fdset].events |=
								POLLOUT;
			io->desc[fd]->events |= EVENT_WRITE;
		}
	} else {
		if (io->desc[fd]->events & EVENT_WRITE) {
			io->fdset[io->desc[fd]->idx_to_fdset].events &=
								~POLLOUT;
			io->desc[fd]->events &= ~EVENT_WRITE;
		}
	}
}

void io_del(struct io_context *io, int fd)
{
	if (!io)
		xerror(1, "io_del: io NULL");

	if (fd < 0)
		xerror(1, "io_del: invalid argument fd=%d", fd);

	if (io->desc[fd] == NULL)
		xerror(1, "io_del: fd=%d is not in the set", fd);

#ifdef IO_PARANOIA
	check_fd(io, fd);
#endif
	/* mark descriptor as deleted */
	io->desc[fd]->deleted_flag = 1;
	io->fd_deleted_flag = 1;
}

int io_loop_once(struct io_context *io)
{
	int res;
	size_t i;
	int timeout;
	short events;
	struct fd_desc *q;

	if (io->fd_deleted_flag) {
		int deleted = 0;
		size_t from, to;
		for (from = to = 0; from < io->nfds_used; from++) {
#ifdef IO_PARANOIA
			check_fd(io, io->fdset[from].fd); /* desc must exist! */
#endif
			if (io->desc[io->fdset[from].fd]->deleted_flag) {
				deleted++;
				xfree(io->desc[io->fdset[from].fd]);
				io->desc[io->fdset[from].fd] = NULL;
				continue;
			}
			if (to != from) {
				/* shift fdset one position left */
				io->fdset[to] = io->fdset[from];
				/* update index in associated desc */
				io->desc[io->fdset[from].fd]->idx_to_fdset -= deleted;
			}
			to++;
		}
		io->nfds_used -= deleted;
		io->fd_deleted_flag = 0;
		if (io->reduse_mem_flag &&
		    (io->nfds_size - io->nfds_used >= (IO_START_SIZE*2))) {
			size_t new_size;
			new_size = ROUND_UP(io->nfds_used, IO_START_SIZE) *
							sizeof(struct pollfd);
			io->fdset = xrealloc0(io->fdset,
					io->nfds_size * sizeof(struct pollfd),
					new_size);
			io->nfds_size = new_size;
		}
	}

	timeout = timeout_next();

	res = poll(io->fdset, io->nfds_used, timeout >= 0 ? timeout : -1);

	if (res >= 0) {

		timeout_run();

		/*
		 * Check loop_running flag because handler can
		 * switch the loop off.
		 */
		for (i = 0; i < io->nfds_used && io->loop_running; i++) {
#ifdef IO_PARANOIA
			check_fd(io, io->fdset[i].fd); /* Be paranoid... */
#endif
			q = io->desc[io->fdset[i].fd];

			if (q->deleted_flag)
				continue;

			if (io->fdset[i].revents) {
				events = 0;

				if (io->fdset[i].revents & (POLLIN|POLLPRI))
					events |= EVENT_READ;
				if (io->fdset[i].revents & POLLOUT) 
					events |= EVENT_WRITE;
				if (io->fdset[i].revents & (POLLHUP|POLLERR|POLLNVAL))
					events |= EVENT_ERROR;
				io->fdset[i].revents = 0;
				io->current_desc = io->fdset[i].fd;
				assert(q->callback != NULL);
				(*q->callback)(io->fdset[i].fd, events, q->opaque);
				io->current_desc = -1;
#ifdef IO_PARANOIA
				check_fd(io, io->fdset[i].fd); /* Yeah, paranoia */
#endif
			}
		}
	} else if (res < 0 && errno != EINTR) {
		xerror(1, "io_loop_once: poll() returned %d: %s",
				res, strerror(errno));
	}
	return 0;
}

void io_loop_stop(struct io_context *io)
{
	io->loop_running = 0;
}

void io_loop_run(struct io_context *io)
{
	if (!io->loop_running)
		io->loop_running = 1;

	while (io->loop_running)
		io_loop_once(io);
}
